class BTree
{
   private String element;
   private BTree left;
   private BTree right;

   public BTree(String element)
   {
      this.element = element;
      left  = null;
      right = null;
   }

   public BTree(String element, BTree left, BTree right)
   {
      this.element = element;
      this.left  = left;
      this.right = right;
   }

   public String getElement()
   {
      return element;
   }

   public BTree getLeftTree()
   {
      return left;
   }

   public BTree getRightTree()
   {
      return right;
   }

   /* Compute the number of subtrees below this node. */
   public int degree()
   {
      int result = 0;
      if (  left != null ) result++;
      if ( right != null ) result++;
      return result;
   }

   /* Calculate the depth of the tree below this node. */
   public int depth()
   {
      int result = 0;
      int depthL = 0;
      if (left != null)
      {
         depthL = getLeftTree().depth();
         result = 1;
      }
      int depthR = 0;
      if (right != null)
      {
         depthR = getRightTree().depth();
         result = 1;
      }
      if (depthL > depthR)
      {
         result += depthL;
      }
      else
      {
         result += depthR;
      }
      return result;
   }

   /* toString() is essentially a preorder traversal of the tree */
   public String toString()
   {
      String result = "";

      if (depth() == 0)  // depth==0 stops the recursion also
      {
         result += getElement();
      }
      else  // do a preorder traversal
      {
         result += "(" + getElement();
         result += " " + ((left!=null)  ? left  : "()");
         result += " " + ((right!=null) ? right : "()") + ")";
      }
      return result;
   }
}//BTree


/**
   The eval() method takes a BTree that is an expression tree
   and evaluates it to a double value.

   The expression tree can contain the following five binary operators,
      +, -, *, /, ^,  and two unary operators neg, sqrt.
   The unary operators have their operand in their right subtree.
*/
class Evaluate
{  /* eval() is essentially a postorder traversal of the tree */
   public static double eval(BTree tree)
   {
      double result = 0.0;

      if (tree.depth() == 0)  // a number (stops the recursion)
      {
         result = Double.parseDouble( tree.getElement() );
      }
      else if ( tree.getElement().equals("S") )  // a unary operator
      {
         result = Math.sqrt( eval(tree.getRightTree()) );
      }
      else if ( tree.getElement().equals("~") )  // a unary operator
      {
         result = -eval(tree.getRightTree());
      }
      else
      {
         double resultL = eval(tree.getLeftTree());
         double resultR = eval(tree.getRightTree());
         String op = tree.getElement();
         if ( op.equals("+") )
         {
            result = resultL + resultR;
         }
         else if ( op.equals("-") )
         {
            result = resultL - resultR;
         }
         else if ( op.equals("*") )
         {
            result = resultL * resultR;
         }
         else if ( op.equals("/") )
         {
            result = resultL / resultR;
         }
         else if ( op.equals("^") )
         {
            result = Math.pow(resultL, resultR);
         }
         else
         {
            throw new Error("Illegal operator: " + op);
         }
      }
      return result;
   }//eval()
}//Evaluate


/**
   The static method parsePrefix() is a recursive descent parser for
   strings that hold the prefix notation of an arithmetic expression.

   Here is the BNF grammar for prefix notation expressions.

      Prefix -> BinOp Prefix Prefix
              | UnOp Prefix
              | Number

       BinOp -> '+' | '-' | '*' | '/'

       UnOp  -> '~' | 'S'

      Number -> any valid Java double

   This parser creates an expression tree (that is, an "abstract syntax tree")
   for each prefix expression.
*/
class ParsePrefix
{
   public static BTree parsePrefix(String expr)
   {
      Tokenizer tokens = new Tokenizer(expr);

      BTree result = getPrefix(tokens); // parse the token stream

      if ( tokens.hasToken() )  // there shouldn't be any more tokens
         throw new Error("Syntax error: unexpected input: "+tokens.peekToken()+"\n"+tokens);

      return result;
   }

   private static BTree getPrefix(Tokenizer tokens)
   {
      BTree result = null;

      String token = tokens.peekToken();  // one token look ahead

      if ( "+-*/^".indexOf(token) != -1 ) // binary operator case
      {
         result = new BTree( tokens.nextToken(),  // consume one token
                             getPrefix(tokens),
                             getPrefix(tokens) );
      }
      else if ( "~S".indexOf(token) != -1 )       // unary operator case
      {
         result = new BTree( tokens.nextToken(),  // consume one token
                             null,
                             getPrefix(tokens) );
      }
      else // operand case (this case terminates the recursion)
      {
         result = new BTree( tokens.nextToken() ); // consume one token
      }
      return result;
   }
}//ParsePrefix


class Tokenizer
{
   private String[] tokens;
   private int currentToken;
   /**
     This constructor converts an array of char (i.e., a String) into an
     array of tokens. Whitespace characters are used as token delimiters.
   */
   public Tokenizer(String string)
   {
      tokens = string.trim().split("\\s+"); // create array of tokens
      currentToken = 0;
   }
   public String peekToken(){ return tokens[currentToken]; }

   public String nextToken(){ return tokens[currentToken++]; }

   public boolean hasToken(){ return currentToken < tokens.length; }
}//Tokenizer


public class JavaProject
{  public static void main(String[] args)
   {                   // prefix notation    // parenthesized infix notation
      String prefix[] = { "+ 2 * 3 4",       // 2 + 3 * 4
                          "* + 2 3 4",       // (2 + 3) * 4
                          "+ - 2 3 4",       // 2 - 3 + 4
                          "- 2 + 3 4",       // 2 - (3 + 4)
                          "~ ^ 2 3",         // -2^3
                          "^ ~ 2 3",         // (-2)^3
                          "/ + ~ -1 S - ^ -1 2 * * 4 1 -1 * 2 1"
                        };
      for (int i = 0; i < prefix.length; i++)
      {
         BTree expTree  = ParsePrefix.parsePrefix( prefix[i] );
         System.out.println("     Original expr --> " + prefix[i]);
         System.out.println("Value of this expr ==> " + Evaluate.eval(expTree)+"\n");
      }
   }//main()
}//JavaProject