(*
   This ML version of the project is more like a
   translation of the Java version into ML than
   it is like a "native" ML version.
*)

datatype btree = Leaf of string
               | Node of string * btree * btree;


exception Evaluate; (* used in the next function *)

fun evalbtree (Leaf s) = valOf(Real.fromString(s))
 |  evalbtree (Node (s, t1, t2)) =
      if s = "Neg" then
        0.0 - evalbtree(t2)
      else if s = "Sqrt" then
        Math.sqrt(evalbtree(t2))
      else if s = "+" then
        evalbtree(t1) + evalbtree(t2)
      else if s = "-" then
        evalbtree(t1) - evalbtree(t2)
      else if s = "*" then
        evalbtree(t1) * evalbtree(t2)
      else if s = "/" then
        evalbtree(t1) / evalbtree(t2)
      else if s = "^" then
        Math.pow(evalbtree(t1), evalbtree(t2))
      else
        raise Evaluate; (* We need this exception because the type
                           system cannot control the contents of a btree. *)


exception Parse; (* used in the next function *)

fun parse tokens =
let
  fun getPrefix [] = raise Parse  (* missing operand (not enough tokens) *)
   |  getPrefix (tk::tks) =
        if isSome(Real.fromString(tk)) then
          (tks, Leaf(tk))
        else
          let val (tk2, t1) = getPrefix tks in
            case tk of
              "~" => (tk2, Node("Neg",  Leaf(""), t1))
            | "S" => (tk2, Node("Sqrt", Leaf(""), t1))
            |  _  =>
              let val (tk3, t2) = getPrefix tk2 in
                case tk of
                  "+" => (tk3, Node("+", t1, t2))
                | "-" => (tk3, Node("-",  t1, t2))
                | "*" => (tk3, Node("*", t1, t2))
                | "/" => (tk3, Node("/",  t1, t2))
                | "^" => (tk3, Node("^",  t1, t2))
                |  _  => raise Parse (* illegal operator *)
              end
          end
in
  let val (tks, etree) = getPrefix(tokens) in
    if not (null tks) then
       raise Parse  (* too many tokens *)
    else
      etree
  end
end;


fun tokenize exprString =
let
  (* tokenizer converts a list of char into a list of tokens.
     Whitespace characters are used as token delimiters. *)
  fun tokenizer [] tk = if tk="" then [] else [tk]
   |  tokenizer (c::cs) tk =
        if c = #" " then
          if tk = "" then
            tokenizer cs tk         (* ignore extra spaces *)
          else
            tk::tokenizer cs ""     (* put a new token in the output list *)
        else
          tokenizer cs (tk^(str c)) (* accumulate a token in the 2nd parameter *)
in
  tokenizer (explode exprString) ""
end;


val prefix = [
(* prefix *)        (* infix *)
"+ 2 * 3 4",        (* 2 + 3 * 4   *)
"* + 2 3 4",        (* (2 + 3) * 4 *)
"+ - 2 3 4",        (* 2 - 3 + 4   *)
"- 2 + 3 4",        (* 2 - (3 + 4) *)
"~ ^ 2 3",          (* -2^3        *)
"^ ~ 2 3",          (* (-2)^3      *)
"/ + ~ -1 S - ^ -1 2 * * 4 1 -1 * 2 1"
];


fun loop [] = []
 |  loop (pf::pfs) = (evalbtree (parse (tokenize pf)))::(loop pfs);

loop prefix;