datatype exprtree = Value of real
                  | Neg  of exprtree
                  | Sqrt of exprtree
                  | Plus of exprtree * exprtree
                  | Sub  of exprtree * exprtree
                  | Prod of exprtree * exprtree
                  | Div  of exprtree * exprtree
                  | Exp  of exprtree * exprtree;


fun evalexpr (Value x) = x
 |  evalexpr (Neg t1) = 0.0 - evalexpr(t1)
 |  evalexpr (Sqrt t1) = Math.sqrt(evalexpr(t1))
 |  evalexpr (Plus (t1,t2)) = evalexpr(t1) + evalexpr(t2)
 |  evalexpr (Sub  (t1,t2)) = evalexpr(t1) - evalexpr(t2)
 |  evalexpr (Prod (t1,t2)) = evalexpr(t1) * evalexpr(t2)
 |  evalexpr (Div  (t1,t2)) = evalexpr(t1) / evalexpr(t2)
 |  evalexpr (Exp  (t1,t2)) = Math.pow(evalexpr(t1), evalexpr(t2));


exception Parse; (* used in the next function *)

fun parse tokens =
let
  fun getPrefix [] = raise Parse (* missing operand (not enough tokens) *)
   |  getPrefix (tk::tks) =
        if isSome(Real.fromString(tk)) then (* check if token is a real *)
          (tks, Value(valOf(Real.fromString(tk))))
        else (* token should be an operator *)
          let val (tk2, t1) = getPrefix tks in
            case tk of
              "~" => (tk2, Neg (t1))
            | "S" => (tk2, Sqrt(t1))
            |  _  =>
              let val (tk3, t2) = getPrefix tk2 in
                case tk of
                  "+" => (tk3, Plus (t1,t2))
                | "-" => (tk3, Sub  (t1,t2))
                | "*" => (tk3, Prod (t1,t2))
                | "/" => (tk3, Div  (t1,t2))
                | "^" => (tk3, Exp  (t1,t2))
                |  _  => raise Parse (* illegal operator *)
              end
          end
in
  let val (tks, etree) = getPrefix(tokens) in
    if not (null tks) then
       raise Parse  (* too many tokens *)
    else
      etree
  end
end;


fun tokenize exprString =
let
  (* tokenizer converts a list of char into a list of tokens.
     Whitespace characters are used as token delimiters. *)
  fun tokenizer [] tk = if tk="" then [] else [tk]
   |  tokenizer (c::cs) tk =
        if c = #" " then
          if tk = "" then
            tokenizer cs tk         (* ignore extra spaces *)
          else
            tk::tokenizer cs ""     (* put a new token in the output list *)
        else
          tokenizer cs (tk^(str c)) (* accumulate a token in the 2nd parameter *)
in
  tokenizer (explode exprString) ""
end;


val prefix = [
(* prefix *)        (* infix *)
"+ 2 * 3 4",        (* 2 + 3 * 4   *)
"* + 2 3 4",        (* (2 + 3) * 4 *)
"+ - 2 3 4",        (* 2 - 3 + 4   *)
"- 2 + 3 4",        (* 2 - (3 + 4) *)
"~ ^ 2 3",          (* -2^3        *)
"^ ~ 2 3",          (* (-2)^3      *)
"/ + ~ -1 S - ^ -1 2 * * 4 1 -1 * 2 1"
];


fun loop [] = []
 |  loop (pf::pfs) = (evalexpr (parse (tokenize pf)))::(loop pfs);

loop prefix;